=== modified file 'bzrlib/tests/__init__.py'
--- bzrlib/tests/__init__.py	2006-04-05 00:59:14 +0000
+++ bzrlib/tests/__init__.py	2006-04-08 16:58:12 +0000
@@ -972,6 +972,7 @@
                    'bzrlib.tests.test_store',
                    'bzrlib.tests.test_symbol_versioning',
                    'bzrlib.tests.test_testament',
+                   'bzrlib.tests.test_textmerge',
                    'bzrlib.tests.test_trace',
                    'bzrlib.tests.test_transactions',
                    'bzrlib.tests.test_transform',

=== added file 'bzrlib/tests/test_textmerge.py'
--- bzrlib/tests/test_textmerge.py	1970-01-01 00:00:00 +0000
+++ bzrlib/tests/test_textmerge.py	2006-04-08 16:58:12 +0000
@@ -0,0 +1,32 @@
+# Copyright (C) 2006 Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+#
+# Author: Aaron Bentley <aaron.bentley@utoronto.ca>
+from bzrlib.textmerge import Merge2
+from bzrlib.tests import TestCase
+class TestMerge2(TestCase):
+    def test_agreed(self):
+        lines = "a\nb\nc\nd\ne\nf\n".splitlines(True)
+        mlines = list(Merge2(lines, lines).merge_lines())
+        self.assertEqualDiff(mlines, lines)
+
+    def test_conflict(self):
+        lines_a = "a\nb\nc\nd\ne\nf\ng\nh\n".splitlines(True)
+        lines_b = "z\nb\nx\nd\ne\ne\nf\ng\ny\n".splitlines(True)
+        expected = "<\na\n=\nz\n>\nb\n<\nc\n=\nx\n>\nd\n<\n=\ne\n>\ne\nf\n"\
+                   "g\n<\nh\n=\ny\n>\n"
+        mlines = Merge2(lines_a, lines_b, '<\n', '>\n', '=\n').merge_lines()
+        self.assertEqualDiff(''.join(mlines), expected)

=== added file 'bzrlib/textmerge.py'
--- bzrlib/textmerge.py	1970-01-01 00:00:00 +0000
+++ bzrlib/textmerge.py	2006-04-08 16:58:12 +0000
@@ -0,0 +1,83 @@
+# Copyright (C) 2005, 2006 Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+#
+# Author: Martin Pool <mbp@canonical.com> 
+#         Aaron Bentley <aaron.bentley@utoronto.ca>
+
+from difflib import SequenceMatcher
+
+
+class TextMerge(object):
+    def __init__(self, a_marker='<<<<<<< \n', b_marker='>>>>>>> \n',
+                 split_marker='=======\n'):
+        self.a_marker = a_marker
+        self.b_marker = b_marker
+        self.split_marker = split_marker
+
+    def struct_to_lines(self, struct_iter):
+        for lines in struct_iter:
+            if len(lines) == 1:
+                for line in lines[0]:
+                    yield line
+            else:
+                yield self.a_marker
+                for line in lines[0]: 
+                    yield line
+                yield self.split_marker
+                for line in lines[1]: 
+                    yield line
+                yield self.b_marker
+
+    def iter_useful(self, struct_iter):
+        for group in struct_iter:
+            if len(group[0]) > 0:
+                yield group
+            elif len(group) > 1 and len(group[1]) > 0:
+                yield group
+
+    def merge_lines(self):
+        return self.struct_to_lines(self.merge_struct())
+
+    def merge_struct(self):
+        return  self.iter_useful(self._merge_struct())
+
+
+class Merge2(TextMerge):
+
+    """
+    Two-way merge.
+    In a two way merge, common regions are shown as unconflicting, and uncommon
+    regions produce conflicts.
+    """
+    def __init__(self, lines_a, lines_b, a_marker='<<<<<<< \n', 
+                 b_marker='>>>>>>> \n', split_marker='=======\n'):
+        TextMerge.__init__(self, a_marker, b_marker, split_marker)
+        self.lines_a = lines_a
+        self.lines_b = lines_b
+
+    def _merge_struct(self):
+        sm = SequenceMatcher(None, self.lines_a, self.lines_b)
+        pos_a = 0
+        pos_b = 0
+        for ai, bi, l in sm.get_matching_blocks():
+            # non-matching lines
+            yield(self.lines_a[pos_a:ai], self.lines_b[pos_b:bi])
+            # matching lines
+            yield(self.lines_a[ai:ai+l],)
+            pos_a = ai + l 
+            pos_b = bi + l
+        # final non-matching lines
+        yield(self.lines_a[pos_a:-1], self.lines_b[pos_b:-1])

=== modified file 'bzrlib/versionedfile.py'
--- bzrlib/versionedfile.py	2006-04-07 23:32:11 +0000
+++ bzrlib/versionedfile.py	2006-04-08 16:58:12 +0000
@@ -31,6 +31,7 @@
 import bzrlib.errors as errors
 from bzrlib.inter import InterObject
 from bzrlib.symbol_versioning import *
+from bzrlib.textmerge import TextMerge
 from bzrlib.transport.memory import MemoryTransport
 from bzrlib.tsort import topo_sort
 from bzrlib import ui
@@ -440,23 +441,14 @@
         yield 'unchanged', ''           # terminator
 
     def weave_merge(self, plan, a_marker='<<<<<<< \n', b_marker='>>>>>>> \n'):
-        return self.struct_to_lines(self.weave_merge_struct(plan), a_marker,
-                                    b_marker)
-
-    def struct_to_lines(self, struct_iter, a_marker, b_marker):
-        for lines in struct_iter:
-            if len(lines) == 1:
-                for line in lines[0]:
-                    yield line
-            else:
-                    yield a_marker
-                    for l in lines[0]: yield l
-                    yield '=======\n'
-                    for l in lines[1]: yield l
-                    yield b_marker
-                    
-
-    def weave_merge_struct(self, plan):
+        return WeaveMerge(plan, a_marker, b_marker).merge_lines()
+
+class WeaveMerge(TextMerge):
+    def __init__(self, plan, a_marker='<<<<<<< \n', b_marker='>>>>>>> \n'):
+        TextMerge.__init__(self, a_marker, b_marker)
+        self.plan = plan
+
+    def _merge_struct(self):
         lines_a = []
         lines_b = []
         ch_a = ch_b = False
@@ -466,7 +458,7 @@
         # We previously considered either 'unchanged' or 'killed-both' lines
         # to be possible places to resynchronize.  However, assuming agreement
         # on killed-both lines may be too agressive. -- mbp 20060324
-        for state, line in plan:
+        for state, line in self.plan:
             if state == 'unchanged':
                 # resync and flush queued conflicts changes if any
                 if not lines_a and not lines_b:

